#ifndef AOCOMMON_IO_SERIAL_STREAM_FWD_H_
#define AOCOMMON_IO_SERIAL_STREAM_FWD_H_

// This header contains the forward declarations for the streaming classes.

namespace aocommon {
class SerialIStream;
class SerialOStream;
}  // namespace aocommon

#endif