#ifndef PREDICT_RESIDUAL_H_
#define PREDICT_RESIDUAL_H_

#include "../../structures/timefrequencydata.h"

namespace algorithms {

void PredictResidual(TimeFrequencyData& data);

}  // namespace algorithms

#endif
