#ifndef AOFLAGGER_RFIGUI_TWO_NUMBERS_DIALOG_H_
#define AOFLAGGER_RFIGUI_TWO_NUMBERS_DIALOG_H_

#include <string>

#include <gtkmm/box.h>
#include <gtkmm/dialog.h>
#include <gtkmm/entry.h>
#include <gtkmm/label.h>
#include <gtkmm/messagedialog.h>

class RangeInputDialog final : public Gtk::MessageDialog {
 public:
  RangeInputDialog(Gtk::Window& parent, const std::string& question,
                   double default_start, double default_end);

  double GetStart() const;
  double GetEnd() const;

 private:
  Gtk::Box box_;
  Gtk::Entry start_range_;
  Gtk::Entry end_range_;
  Gtk::Label dash_label_{"-"};
};

#endif
