#ifndef AOCOMMON_CONSTANTS_H_
#define AOCOMMON_CONSTANTS_H_

namespace aocommon {

/**
 * The speed of light in meters/second.
 */
constexpr double kSpeedOfLight = 299792458.0;

}  // namespace aocommon

#endif
